Package: transformr
Type: Package
Title: Polygon and Path Transformations
Version: 0.1.5
Authors@R: 
    person(given = "Thomas Lin",
           family = "Pedersen",
           role = c("cre", "aut"),
           email = "thomasp85@gmail.com",
           comment = c(ORCID = "0000-0002-5147-4711"))
Maintainer: Thomas Lin Pedersen <thomasp85@gmail.com>
Description: In order to smoothly animate the transformation of polygons and 
    paths, many aspects needs to be taken into account, such as differing number 
    of control points, changing center of rotation, etc. The 'transformr' 
    package provides an extensive framework for manipulating the shapes of 
    polygons and paths and can be seen as the spatial brother to the 'tweenr' 
    package.
License: MIT + file LICENSE
Encoding: UTF-8
Imports: tweenr, rlang, sf, lpSolve, vctrs
Suggests: covr, magrittr
LinkingTo: cpp11
RoxygenNote: 7.3.1
URL: https://github.com/thomasp85/transformr
BugReports: https://github.com/thomasp85/transformr/issues
NeedsCompilation: yes
Packaged: 2024-02-26 13:46:42 UTC; thomas
Author: Thomas Lin Pedersen [cre, aut]
    (<https://orcid.org/0000-0002-5147-4711>)
Repository: CRAN
Date/Publication: 2024-02-26 14:30:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-26 05:33:25 UTC; unix
Archs: transformr.so.dSYM

# Generated by roxygen2: do not edit by hand

export(path_spiral)
export(path_waves)
export(point_grid)
export(point_random)
export(poly_circle)
export(poly_circles)
export(poly_star)
export(poly_star_hole)
export(st_normalize)
export(tween_path)
export(tween_polygon)
export(tween_sf)
import(vctrs)
importFrom(lpSolve,lp.assign)
importFrom(rlang,"%||%")
importFrom(rlang,as_function)
importFrom(rlang,enquo)
importFrom(rlang,eval_tidy)
importFrom(rlang,quo)
importFrom(rlang,quo_is_null)
importFrom(sf,"st_crs<-")
importFrom(sf,"st_geometry<-")
importFrom(sf,st_area)
importFrom(sf,st_as_sfc)
importFrom(sf,st_bbox)
importFrom(sf,st_cast)
importFrom(sf,st_centroid)
importFrom(sf,st_combine)
importFrom(sf,st_crs)
importFrom(sf,st_distance)
importFrom(sf,st_geometry)
importFrom(sf,st_geometry_type)
importFrom(sf,st_intersection)
importFrom(sf,st_length)
importFrom(sf,st_linestring)
importFrom(sf,st_multilinestring)
importFrom(sf,st_multipoint)
importFrom(sf,st_multipolygon)
importFrom(sf,st_point)
importFrom(sf,st_polygon)
importFrom(sf,st_sample)
importFrom(sf,st_sfc)
importFrom(sf,st_touches)
importFrom(sf,st_transform)
importFrom(sf,st_union)
importFrom(sf,st_voronoi)
importFrom(stats,runif)
importFrom(tweenr,.complete_states)
importFrom(tweenr,.get_last_frame)
importFrom(tweenr,.has_frames)
importFrom(tweenr,.max_id)
importFrom(tweenr,.with_prior_frames)
importFrom(tweenr,tween_state)
useDynLib(transformr, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# transformr <img src="man/figures/logo.png" align="right"/>

<!-- badges: start -->

[![R-CMD-check](https://github.com/thomasp85/transformr/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/thomasp85/transformr/actions/workflows/R-CMD-check.yaml)
[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version-ago/transformr)](https://cran.r-project.org/package=transformr)
[![CRAN_Download_Badge](https://cranlogs.r-pkg.org/badges/grand-total/transformr)](https://cran.r-project.org/package=transformr)
<!-- badges: end -->

If you’ve ever made animated data visualisations you’ll know that
arbitrary polygons and lines requires special considerations if the
animation is to be smooth and believable. `transformr` is able to remove
all of these worries by expanding
[`tweenr`](https://github.com/thomasp85/tweenr) to understand spatial
data, and thus lets you focus on defining your animation steps.
`transformr` takes care of matching shapes between states, cutting some
in bits if the number doesn’t match between the states, and ensures that
each pair of matched shapes contains the same number of anchor points
and that these are paired up so as to avoid rotation and inversion
during animation.

`transformr` supports both polygons (with holes), and paths either
encoded as simple x/y data.frames or as simpel features using the
[`sf`](https://github.com/r-spatial/sf) package.

## Installation

You can install transformr from CRAN using
`install.packages('transformr')` or grab the development version from
github with:

``` r
# install.packages("devtools")
devtools::install_github("thomasp85/transformr")
```

## Examples

These are simple, contrieved examples showing how the API works. It
scales simply to more complicated shapes.

### Polygon

A polygon is simply a data.frame with an `x` and `y` column, where each
row demarcates an anchor point for the polygon. The polygon is not in
closed form, that is, the first point is not repeated in the end. If
more polygons are wanted you can provide an additional column that
indicate the polygon membership of a column (quite like
`ggplot2::geom_polygon()` expects an `x`, `y`, and `group` variable). If
holed polygons are needed, holes should follow the main polygon and be
separated with an `NA` row in the `x` and `y` column.

``` r
library(transformr)
library(tweenr)
library(ggplot2)
polyplot <- function(data) {
  p <- ggplot(data) + 
    geom_polygon(aes(x, y, group = id, fill = col)) +
    scale_fill_identity() +
    coord_fixed(xlim = c(-1.5, 1.5), ylim = c(-1.5, 1.5))
  plot(p)
}

star <- poly_star()
star$col <- 'steelblue'
circles <- poly_circles()
circles$col <- c('forestgreen', 'firebrick', 'goldenrod')[circles$id]

animation <- tween_polygon(star, circles, 'cubic-in-out', 40, id) %>% 
  keep_state(10)

ani <- lapply(split(animation, animation$.frame), polyplot)
```

![](man/figures/README-unnamed-chunk-2.gif)

By default the polygons are matched up based on their id. In the above
example there’s a lack of polygons in the start-state, so these have to
appear somehow. This is governed by the `enter` function, which by
default is `NULL` meaning new polygons just appear at the end of the
animation. We can change this to get a nicer result:

``` r
# Make new polygons appear 2 units below their end position
from_below <- function(data) {
  data$y <- data$y - 2
  data
}
animation <- tween_polygon(star, circles, 'cubic-in-out', 40, id, enter = from_below) %>% 
  keep_state(10)

ani <- lapply(split(animation, animation$.frame), polyplot)
```

![](man/figures/README-unnamed-chunk-3.gif)

Similar to the `enter` function it is possible to supply an `exit`
function when the start state has more polygons than the end state.
These functions get a single polygon with the state it was/will be, that
can then be manipulated at will, as long as the same number of rows and
columns are returned.

> The `enter` and `exit` functions have slightly different semantics
> here than in `tweenr::tween_state()` where it gets all
> entering/exiting rows in one go, and not one-by-one

Our last option is to not match the polygons up, but simply say “make
everything in the first state, into everything in the last state…
somehow”. This involves cutting up polygons in the state with fewest
polygons and match polygons by minimizing the distance and area
difference between pairs. All of this is controlled by setting
`match = FALSE` in `tween_polygon()`, and `transformr` will then do its
magic:

``` r
animation <- tween_polygon(star, circles, 'cubic-in-out', 40, id, match = FALSE) %>% 
  keep_state(10)

ani <- lapply(split(animation, animation$.frame), polyplot)
```

![](man/figures/README-unnamed-chunk-4.gif)

### Paths

Paths are a lot like polygons, except that they don’t *wrap-around*.
Still, slight differences in how they are tweened exists. Chief among
these are that the winding order are not changed to minimize the
travel-distance, because paths often have an implicit direction and this
should not be tampered with. Further, when automatic matching paths
(that is, `match = FALSE`), paths are matched to minimize the difference
in length as well as the pair distance. The same interpretation of the
`enter`, `exit`, and `match` arguments remain, which can be seen in the
two examples below:

``` r
pathplot <- function(data) {
  p <- ggplot(data) + 
    geom_path(aes(x, y, group = id)) +
    coord_fixed(xlim = c(-1.5, 1.5), ylim = c(-1.5, 1.5))
  plot(p)
}
spiral <- path_spiral()
waves <- path_waves()

animation <- tween_path(spiral, waves, 'cubic-in-out', 40, id, enter = from_below) %>% 
  keep_state(10)

ani <- lapply(split(animation, animation$.frame), pathplot)
```

![](man/figures/README-unnamed-chunk-5.gif)

``` r
animation <- tween_path(spiral, waves, 'cubic-in-out', 40, id, match = FALSE) %>% 
  keep_state(10)

ani <- lapply(split(animation, animation$.frame), pathplot)
```

![](man/figures/README-unnamed-chunk-6.gif)

### Simple features

The `sf` package provides an implemention of simple features which are a
way to encode any type of geometry in defined classes and operate on
them. `transformr` supports (multi)point, (multi)linestring, and
(multi)polygon geometries which acount for most of the use cases. When
using the `tween_sf()` function any `sfc` column will be tweened by
itself, while the rest will be tweened by `tweenr::tween_state()`. For
any *multi* type, the tweening progress as if `match = FALSE` in
`tween_polygon()` and `tween_path()`, that is polygons/paths are cut and
matched to even out the two states. For multipoint the most central
points are replicated to ensure the same number of points in each state.
One nice thing about `sf` is that you can encode different geometry
types in the same data.frame and plot it all at once:

``` r
sfplot <- function(data) {
  p <- ggplot(data) + 
    geom_sf(aes(colour = col, geometry = geometry)) + 
    coord_sf(datum = NA) + # remove graticule
    scale_colour_identity()
  plot(p)
}
star_hole <- poly_star_hole(st = TRUE)
circles <- poly_circles(st = TRUE)
spiral <- path_spiral(st = TRUE)
waves <- path_waves(st = TRUE)
random <- point_random(st = TRUE)
grid <- point_grid(st = TRUE)
df1 <- data.frame(
  geo = sf::st_sfc(star_hole, spiral, random),
  col = c('steelblue', 'forestgreen', 'goldenrod')
)
df2 <- data.frame(
  geo = sf::st_sfc(circles, waves, grid),
  col = c('goldenrod', 'firebrick', 'steelblue')
)

animation <- tween_sf(df1, df2, 'cubic-in-out', 40) %>% 
  keep_state(10)

ani <- lapply(split(animation, animation$.frame), sfplot)
```

![](man/figures/README-unnamed-chunk-7.gif)

#' Transition between polygon data.frames
#'
#' This function is equivalent to [tweenr::tween_state()] except that data is
#' interpeted as encoding polygons. Data is expected to have an `x` and `y`
#' column encoding the location of corners in the polygon.
#'
#' @inheritParams tweenr::tween_state
#' @param match Should polygons be matched by id? If `FALSE` then polygons will
#' be matched by shortest distance and if any state has more polygons than the
#' other, the other states polygons will be chopped up so the numbers match.
#'
#' @return A data.frame containing intermediary states
#'
#' @section Aligning polygons:
#' `transformr` performs a lot of work to try to ensure the transition between
#' different shapes are as smooth and direct as possible. The first operation is
#' to ensure that the two end states of the polygon are both drawn clockwise, so
#' that the transition will not contain an inversion. Second, we need to make
#' sure that each end state is drawn with the same number of points. If not, the
#' less detailed polygon will get points inserted at the longest edges so that
#' the number is even between the two states. Third, we rotate the last state so
#' as to minimize the cumulative distance between all point pairs, thus ensuring
#' that the transition will involve a minimum of rotation.
#'
#' @section Cutting polygons:
#' If the transition involves changing the number of polygons, there are two
#' strategies: Making polygons appear/disappear to even out the number, or
#' cutting up the polygons in the state with the fewest in order to create the
#' same number of polygons for the transition. In the latter case, a choice have
#' to be made with regards to which polygons to cut, into how many, and where to
#' cut it. `transformr` will distribute the number of cuts among candidate
#' polygons based on their relative area, ensuring that it is not necessarily
#' the largest polygon that gets all the cuts, but that divisions are
#' distributed as fairly as possible. For deciding on where to cut the polygons
#' they are triangulated and the triangles are then reassembled into the number
#' of pieces needed by always adding to the smallest piece.
#'
#' @section Polygon with holes:
#' `transformr` support polygons with any number of holes. Holes are encoded by
#' adding an `NA` row to the main enclosing polygon and appending the hole after
#' that. Multiple holes are likewise added by simply separating them with `NA`
#' rows. A hole might get cut up and disappear during transition if the polygon
#' needs to be divided. When transitioning between polygons with holes the holes
#' are matched by position to minimize the travel distance. If there is a
#' mismatch between the number of holes in each end state then new zero-area
#' holes are inserted in the centroid of the polygon with the fewest to even out
#' the number.
#'
#' @export
#' @importFrom tweenr tween_state .get_last_frame .with_prior_frames .has_frames
#' @importFrom rlang enquo quo_is_null quo eval_tidy %||%
#'
#' @examples
#' library(magrittr)
#' star <- poly_star_hole()
#' circle <- poly_circle()
#' circles <- poly_circles()
#'
#' tween_polygon(circle, star, 'cubic-in-out', 20) %>%
#'   tween_polygon(circles, 'cubic-in-out', 20)
#'
tween_polygon <- function(.data, to, ease, nframes, id = NULL, enter = NULL, exit = NULL, match = TRUE) {
  if (!is.data.frame(.data)) {
    stop('`.data` must be a `data.frame`', call. = FALSE)
  }
  from <- .get_last_frame(.data)
  id <- enquo(id)
  from$.id <- eval_tidy(id, from) %||% rep(1L, nrow(from))
  from$.phase <- rep('raw', nrow(from))
  to$.id <- eval_tidy(id, to) %||% rep(1L, nrow(to))
  to$.phase <- rep('raw', nrow(to))
  if (.has_frames(.data)) nframes <- nframes + 1
  polygons <- align_polygons(from, to, enter = enter, exit = exit, match = match)
  polygons <- tween_state(polygons$from, polygons$to, ease = ease, nframes = nframes)
  polygons <- polygons[!polygons$.frame %in% c(1, nframes), , drop = FALSE]
  if (is.character(polygons$.id) || is.character(from$.id) || is.character(to$.id)) {
    polygons$.id <- as.character(polygons$.id)
    from$.id <- as.character(from$.id)
    to$.id <- as.character(to$.id)
  }
  morph <- vec_rbind(
    if (nframes > 1) cbind(from, .frame = rep(1, nrow(from))) else NULL,
    polygons,
    cbind(to, .frame = rep(nframes, nrow(to)))
  )
  .with_prior_frames(.data, morph, nframes)
}

align_polygons <- function(from, to, min_n = 50, enter, exit, match = TRUE) {
  from <- lapply(make_polygons(from), function(x) as_clockwise(list(x))[[1]])
  to <- lapply(make_polygons(to), function(x) as_clockwise(list(x))[[1]])
  polygons <- if (match) {
    prep_match_polygons(from, to)
  } else {
    prep_align_polygons(from, to)
  }
  polygons <- mapply(
    match_shapes,
    from = polygons$from,
    to = polygons$to,
    MoreArgs = list(
      enter = enter, exit = exit, min_n = min_n, closed = TRUE
    ),
    SIMPLIFY = FALSE
  )
  from <- lapply(polygons, `[[`, 'from')
  id <- rep(seq_along(from), vapply(from, nrow, integer(1)))
  from <- vec_rbind(!!!from)
  to <- vec_rbind(!!!lapply(polygons, `[[`, 'to'))
  from$.id <- id
  to$.id <- id
  common_id(from = from, to = to)
}

prep_match_polygons <- function(from, to) {
  all_ids <- as.character(union(names(from), names(to)))
  from_all <- structure(rep(list(NULL), length(all_ids)), names = all_ids)
  to_all <- from_all
  from_all[names(from)] <- from
  to_all[names(to)] <- to
  list(from = from_all, to = to_all)
}
#' @importFrom sf st_sfc st_area st_distance st_centroid
#' @importFrom lpSolve lp.assign
prep_align_polygons <- function(from, to) {
  from_st <- st_sfc(lapply(from, to_polygon))
  to_st <- st_sfc(lapply(to, to_polygon))
  if (length(from) < length(to)) {
    area <- st_area(from_st)
    from <- divide_polygons(from, area, length(to) - length(from))
    from_st <- st_sfc(lapply(from, to_polygon))
  } else if (length(to) < length(from)) {
    area <- st_area(to_st)
    to <- divide_polygons(to, area, length(from) - length(to))
    to_st <- st_sfc(lapply(to, to_polygon))
  }
  from_st <- st_normalize(from_st)
  to_st <- st_normalize(to_st)
  if (length(from_st) != 1 && length(to_st) != 1) {
    distance <- st_distance(st_centroid(from_st), st_centroid(to_st))
    area_diff <- abs(outer(st_area(from_st), st_area(to_st), `-`))
    if (max(area_diff) > 0) distance <- distance * (1 + area_diff / max(area_diff))
    match_poly <- lp.assign(distance)
    if (match_poly$status == 0) {
      to <- to[apply(round(match_poly$solution) == 1, 1, which)]
    }
  }
  names(from) <- as.character(seq_along(from))
  names(to) <- as.character(seq_along(to))
  list(from = from, to = to)
}
make_polygons <- function(x) {
  lapply(split(x, x$.id), function(xx) {
    holes <- which(is.na(xx$x))
    if (length(holes) == 0) {
      xx <- list(xx)
    } else {
      holes <- c(holes, nrow(xx) + 1)
      xx <- xx[-holes, , drop = FALSE]
      xx <- split(xx, rep(seq_along(holes), diff(c(0, holes)) - 1))
    }
    lapply(xx, function(ring) {
      nr <- nrow(ring)
      if (nr > 0 && nr < 4) {
        ring[c(seq_len(nr), rep(1, 4 - nr)), ]
      } else {
        ring
      }
    })
  })
}
align_holes <- function(from, to) {
  from_st <- st_sfc(lapply(from, function(x) to_polygon(list(x))))
  to_st <- st_sfc(lapply(to, function(x) to_polygon(list(x))))
  from_centroids <- st_centroid(from_st)
  to_centroids <- st_centroid(to_st)
  n_from <- length(from_st)
  n_to <- length(to_st)
  if (n_from < n_to) {
    from_centroids <- from_centroids[c(seq_len(n_from), rep(1, n_to - n_from))]
  }
  if (n_from > n_to) {
    to_centroids <- to_centroids[c(seq_len(n_to), rep(1, n_from - n_to))]
  }
  distance <- st_distance(from_centroids[-1], to_centroids[-1])
  match_poly <- lp.assign(distance)
  from <- from[seq_along(from_centroids)]
  if (match_poly$status == 0) {
    to <- to[c(1, apply(round(match_poly$solution) == 1, 1, which) + 1)]
  } else {
    to <- to[seq_along(to_centroids)]
  }
  holes <- Map(function(from, to) {
    if (is.null(from)) {
      from <- to
      from$x <- from_centroids[[1]][1]
      from$y <- from_centroids[[1]][2]
    } else if (is.null(to)) {
      to <- from
      to$x <- to_centroids[[1]][1]
      to$y <- to_centroids[[1]][2]
    } else {
      n_points <- max(nrow(from), nrow(to))
      if (nrow(from) < n_points) from <- add_points(from, n_points - nrow(from))
      if (nrow(to) < n_points) main_to <- add_points(to, n_points - nrow(to))
      offset <- rotate(to$x, to$y, from$x, from$y)
      to_end <- seq_len(nrow(to)) < offset
      to <- vec_rbind(to[!to_end, , drop = FALSE],
                      to[to_end, , drop = FALSE])
    }
    list(from = from, to = to)
  }, from = from[-1], to = to[-1])
  list(from = lapply(holes, `[[`, 'from'), to = lapply(holes, `[[`, 'to'))
}
as_clockwise <- function(polygon) {
  lapply(polygon, function(poly) {
    lapply(poly, function(p) {
      x <- c(p$x, p$x[1])
      y <- c(p$y, p$y[1])
      area <- (x[-1] - x[-length(x)]) * (y[-1] + y[-length(y)])
      if (sum(area) < 0) p <- p[rev(seq_len(nrow(p))), ]
      p
    })
  })
}
#' @importFrom sf st_polygon st_sample st_combine st_intersection st_cast st_voronoi st_as_sfc st_bbox st_sfc st_area st_touches st_union
split_polygon <- function(polygon, n) {
  poly <- st_polygon(lapply(polygon, function(p) cbind(p$x, p$y)[c(seq_len(nrow(p)), 1), ]))
  rand <- st_sample(poly, n)
  while (length(rand) < n) {
    rand <- c(rand, st_sample(poly, n))
  }
  rand <- st_combine(rand[sample(length(rand), n)])
  tiles <- st_intersection(st_cast(st_voronoi(rand, st_as_sfc(st_bbox(poly)))), poly)
  tiles <- unlist(lapply(tiles, function(x) {
    if (inherits(x, 'MULTIPOLYGON')) {
      lapply(x, st_polygon)
    } else {
      list(x)
    }
  }), recursive = FALSE)
  while (length(tiles) > n) {
    st_tiles <- st_sfc(tiles)
    smallest <- which.min(st_area(st_tiles))
    touches <- st_touches(st_tiles[smallest], st_tiles)
    neighbor <- touches[[1]][1]
    combined <- st_union(tiles[[smallest]], tiles[[neighbor]])
    tiles <- c(tiles[-c(smallest, neighbor)], list(combined))
  }
  all_points <- vec_rbind(!!!polygon)
  id <- paste0(all_points$x, '-', all_points$y)
  lapply(tiles, function(t) {
    x <- t[[1]][-nrow(t[[1]]), 1]
    y <- t[[1]][-nrow(t[[1]]), 2]
    orig_points <- match(paste0(x, '-' , y), id)
    if (all(is.na(orig_points))) {
      tile <- all_points[rep(1, length(orig_points)), , drop = FALSE]
    } else {
      orig_points <- orig_points[fill_down(is.na(orig_points))]
      tile <- all_points[orig_points, , drop = FALSE]
    }
    tile$x <- x
    tile$y <- y
    list(tile)
  })
}
divide_polygons <- function(polygons, area, n) {
  biggest <- order(area, decreasing = TRUE)
  n_splits <- find_splits(area[biggest], n)
  splits <- n_splits[match(seq_along(area), biggest)] + 1
  unlist(Map(split_polygon, polygon = polygons, n = splits), recursive = FALSE)
}
#' @importFrom sf st_polygon
to_polygon <- function(points) {
  points <- points[[1]]
  if (is.data.frame(points)) {
    points <- as.matrix(points[, c('x', 'y')])
  } else {
    points <- points[, 1:2]
  }
  if (anyNA(points)) {
    first_na <- which(is.na(points[,1]))
    points <- points[seq_len(first_na-1), , drop = FALSE]
  }
  st_polygon(list(rbind(points, points[1,])))
}
#' Normalise a geometry to fit inside a unit square
#'
#' This is a small helper function that will take an sf geometry and fit it
#' inside the unit square (a square centered on 0 and ranging from -1 to 1 in
#' both dimensions). The function will retain the aspect ratio of the geometry
#' and simply scale it down until it fits.
#'
#' @param st An sf geometry such as `sf`, `sfc`, or `sfg`
#'
#' @return An object of the same type as `st`
#'
#' @importFrom sf st_bbox st_geometry st_geometry<-
#' @export
#'
#' @examples
#' library(sf)
#' nc <- st_read(system.file("shape/nc.shp", package="sf"), quiet = TRUE)
#' st_bbox(nc)
#'
#' nc_norm <- st_normalize(nc)
#' st_bbox(nc_norm)
#'
st_normalize <- function(st) {
  if (inherits(st, 'sf')) {
    st_geometry(st) <- st_normalize(st_geometry(st))
    st
  } else {
    bbox <- st_bbox(st)
    x <- bbox$xmin + (bbox$xmax - bbox$xmin) / 2
    y <- bbox$ymin + (bbox$ymax - bbox$ymin) / 2
    st <- st - c(x, y)
    st / (0.5 * max(bbox$xmax - bbox$xmin, bbox$ymax - bbox$ymin))
  }
}

#' Transition between data.frames containing sfc columns
#'
#' This function is equivalent to [tweenr::tween_state()] except that it
#' understands [`sf::sfc`] columns, as defined by the `sf` package. An `sfc`
#' column is a column containing simple features and can this hold both points,
#' lines polygons and more. `tween_sf` currently has support for (multi)point,
#' (multi)linestring, and (multi)polygon types and requires that the transition
#' is between compatible types (points-to-points, linestring-to-linestring,
#' polygon-to-polygon). For (multi)linestring and (multi)polygon, the behavior
#' is similar to [tween_path()] and [tween_polygon()] respectively, with each
#' feature beeing run through the respective function with `match = FALSE`. For
#' (multi)points it behaves more or less like [tweenr::tween_state()] except
#' additional points are added as needed to make the to stages contain the same
#' number of points. Points are added on top of existing points so it appears as
#' if the points are divided into more.
#'
#' @inheritParams tweenr::tween_state
#'
#' @return A data.frame containing intermediary states
#'
#' @importFrom tweenr .complete_states .max_id .has_frames
#' @importFrom rlang enquo
#' @importFrom sf st_crs st_crs<- st_transform
#' @export
#'
#' @examples
#' library(magrittr)
#' star_hole <- poly_star_hole(st = TRUE)
#' circles <- poly_circles(st = TRUE)
#' spiral <- path_spiral(st = TRUE)
#' waves <- path_waves(st = TRUE)
#' random <- point_random(st = TRUE)
#' grid <- point_grid(st = TRUE)
#' df1 <- data.frame(geo = sf::st_sfc(star_hole, spiral, random))
#' df2 <- data.frame(geo = sf::st_sfc(circles, waves, grid))
#'
#' tween_sf(df1, df2, 'linear', 30)
#'
tween_sf <- function(.data, to, ease, nframes, id = NULL, enter = NULL, exit = NULL) {
  if (!is.data.frame(.data)) {
    stop('`.data` must be a `data.frame`', call. = FALSE)
  }
  sf_columns <- vapply(.data, inherits, logical(1), 'sfc')
  if (!any(sf_columns)) return(tween_state(.data, to, ease, nframes, !!id, enter, exit))

  for (col in which(sf_columns)) {
    from_crs <- st_crs(.data[[col]])
    to_crs <- st_crs(to[[col]])
    if (is.na(from_crs)) {
      if (!is.na(to_crs)) {
        st_crs(.data[[col]]) <- to_crs
      }
      next
    }
    if (is.na(to_crs)) {
      st_crs(to[[col]]) <- from_crs
    } else {
      to[[col]] <- st_transform(to[[col]], from_crs)
    }
  }

  from <- .get_last_frame(.data)
  from$.phase <- rep('raw', nrow(from))
  to$.phase <- rep('raw', nrow(to))
  to$.id <- rep(NA_integer_, nrow(to))
  id <- enquo(id)
  if (.has_frames(.data)) nframes <- nframes + 1

  # recalc to make sure .phase and .id columns are included
  sf_columns <- vapply(from, inherits, logical(1), 'sfc')

  full_set <- .complete_states(from, to, id, enter, exit, .max_id(.data))
  to$.id <- full_set$orig_to
  sf_from <- full_set$from[, sf_columns, drop = FALSE]
  sf_to <- full_set$to[, sf_columns, drop = FALSE]
  full_set$from[, sf_columns] <- rep(1L, nrow(full_set$from))
  full_set$to[, sf_columns] <- rep(1L, nrow(full_set$to))
  morph <- tween_state(as.data.frame(full_set$from), as.data.frame(full_set$to), ease, nframes, id = NULL, enter, exit)
  morph[which(sf_columns)] <- tween_sf_col(sf_from, sf_to, rep(ease, length.out = ncol(from))[sf_columns], nframes)
  morph <- morph[!morph$.frame %in% c(1, nframes), , drop = FALSE]
  if (nrow(morph) == 0) morph <- NULL
  morph <- vec_rbind(
    if (nframes > 1) cbind(as.data.frame(from), .frame = rep(1, nrow(from))) else NULL,
    morph,
    cbind(as.data.frame(to), .frame = rep(nframes, nrow(to)))
  )
  .with_prior_frames(.data, morph, nframes)
}
#' @importFrom sf st_geometry_type st_sfc
#' @importFrom tweenr tween_state
tween_sf_col <- function(from, to, ease, nframes) {
  lapply(seq_along(from), function(i) {
    aligned <- align_sf(from[[i]], to[[i]])
    if (is.null(aligned)) return(st_sfc(list()))
    tweened <- tween_state(aligned$from, aligned$to, ease, nframes)
    tweened$id <- as.integer(tweened$id)
    tweened$sf_id <- as.integer(tweened$sf_id)
    tweened$.frame <- as.integer(tweened$.frame)
    st_sfc(repack_sf(tweened, aligned$type, as.integer(nframes)), crs = st_crs(from[[i]]))
  })
}

align_sf <- function(from, to) {
  if (length(from) == 0 && length(to) == 0) return(NULL)
  from_type <- as.character(unlist(lapply(from, st_geometry_type)))
  if (!all(from_type %in% supp_types)) stop('Unsupported geometry type', call. = FALSE)
  to_type <- as.character(unlist(lapply(to, st_geometry_type)))
  if (!all(to_type %in% supp_types)) stop('Unsupported geometry type', call. = FALSE)
  if (!all(sub('MULTI', '', from_type) == sub('MULTI', '', to_type))) stop('Incompatible geometry types', call. = FALSE)
  from <- unpack_sf(from, from_type)
  to <- unpack_sf(to, to_type)
  aligned <- Map(function(from, to, type) {
    switch(
      type,
      POINT =,
      MULTIPOINT = align_sf_point(from, to),
      LINESTRING =,
      MULTILINESTRING = align_sf_path(from, to, 50),
      POLYGON =,
      MULTIPOLYGON = align_sf_polygon(from, to, 50)
    )
  }, from = from, to = to, type = from_type)
  from <- lapply(aligned, `[[`, 'from')
  to <- lapply(aligned, `[[`, 'to')
  id <- rep(seq_along(from), vapply(from, nrow, integer(1)))
  from <- vec_rbind(!!!from)
  to <- vec_rbind(!!!to)
  from$sf_id <- id
  to$sf_id <- id
  list(from = from, to = to, type = from_type)
}
#' @importFrom sf st_point st_distance
#' @importFrom lpSolve lp.assign
align_sf_point <- function(from, to) {
  f_st <- lapply(seq_len(nrow(from)), function(i) st_point(c(from$x[i], from$y[i])))
  t_st <- lapply(seq_len(nrow(to)), function(i) st_point(c(to$x[i], to$y[i])))
  dist <- st_distance(st_sfc(f_st), st_sfc(t_st))
  if (nrow(dist) < ncol(dist)) {
    closest <- apply(dist, 1, mean)
    to_add <- rep(order(closest), length.out = ncol(dist) - nrow(dist))
    dist <- rbind(dist, dist[to_add, , drop = FALSE])
    from <- vec_rbind(from, from[to_add, , drop = FALSE])
  } else if (nrow(dist) > ncol(dist)) {
    closest <- apply(dist, 2, mean)
    to_add <- rep(order(closest), length.out = nrow(dist) - ncol(dist))
    dist <- cbind(dist, dist[, to_add, drop = FALSE])
    to <- vec_rbind(to, to[to_add, , drop = FALSE])
  }
  match_points <- lp.assign(dist)
  if (match_points$status == 0) {
    to <- to[apply(round(match_points$solution) == 1, 1, which), , drop = FALSE]
  }
  from$id <- seq_len(nrow(from))
  to$id <- from$id
  list(from = from, to = to)
}
align_sf_path <- function(from, to, min_n) {
  prepped <- prep_align_paths(from, to)
  polygons <- mapply(
    match_shapes,
    from = prepped$from,
    to = prepped$to,
    MoreArgs = list(
      enter = NULL, exit = NULL, min_n = min_n, closed = FALSE
    ),
    SIMPLIFY = FALSE
  )
  from <- lapply(polygons, `[[`, 'from')
  id <- rep(seq_along(from), vapply(from, nrow, integer(1)))
  from <- vec_rbind(!!!from)
  to <- vec_rbind(!!!lapply(polygons, `[[`, 'to'))
  from$id <- id
  to$id <- id
  list(from = from, to = to)
}
align_sf_polygon <- function(from ,to, min_n) {
  from <- as_clockwise(from)
  to <- as_clockwise(to)
  prepped <- prep_align_polygons(from, to)
  polygons <- mapply(
    match_shapes,
    from = prepped$from,
    to = prepped$to,
    MoreArgs = list(
      enter = NULL, exit = NULL, min_n = min_n, closed = TRUE
    ),
    SIMPLIFY = FALSE
  )
  from <- lapply(polygons, `[[`, 'from')
  id <- rep(seq_along(from), vapply(from, nrow, integer(1)))
  from <- vec_rbind(!!!from)
  to <- vec_rbind(!!!lapply(polygons, `[[`, 'to'))
  from$id <- id
  to$id <- id
  list(from = from, to = to)
}
supp_types <- c('POINT', 'LINESTRING', 'POLYGON', 'MULTIPOINT', 'MULTILINESTRING', 'MULTIPOLYGON')

#' Transition between path data frames
#'
#' This function is equivalent to [tweenr::tween_state()] but expects the data
#' to have an `x` and `y` column and encode paths.
#'
#' @inheritParams tween_polygon
#'
#' @return A data.frame containing intermediary states
#'
#' @section Aligning paths:
#' There is less work required to align paths than there is to align polygons,
#' simply because no rotation is possible/required, and the notion of clockwise
#' winding order is not meaningful in the scope of paths. Still, paths need to
#' be matched and the number of points in each pair of matched paths must be
#' equal. Paths are matched based on relative length rather than on position and
#' seek to minimize the change in length during transition. This is chosen from
#' the point of view that huge elongation or contraction are much more
#' distracting than longer travel distances.
#'
#' @section Cutting paths:
#' If the number of paths to transition between is not even, some of the paths
#' need to be cut in order to succesfully match the paths. The cuts are
#' distributed based on the same algorithm that distributes cuts in polygons and
#' seek to cut the lines into as even-length pieces as possible.
#'
#' @section Multipaths:
#' It is possible to encode multiple paths with the same id be separating them
#' with a `NA` row, much in the same way as holes are encoded in polygons. If
#' paths are not matched based on id (`match = FALSE`) then multipaths will
#' simply be split into separate paths. On the other hand, if paths are matched
#' by id all paths within a multipath will transition into the (multi)path that
#' has the same id in the other state.
#'
#' @importFrom rlang enquo quo_is_null eval_tidy %||%
#' @importFrom tweenr .has_frames tween_state
#' @export
tween_path <- function(.data, to, ease, nframes, id = NULL, enter = NULL, exit = NULL, match = TRUE) {
  if (!is.data.frame(.data)) {
    stop('`.data` must be a `data.frame`', call. = FALSE)
  }
  from <- .get_last_frame(.data)
  id <- enquo(id)
  from$.id <- eval_tidy(id, from) %||% rep(1L, nrow(from))
  from$.phase <- rep('raw', nrow(from))
  to$.id <- eval_tidy(id, to) %||% rep(1L, nrow(to))
  to$.phase <- rep('raw', nrow(to))
  if (.has_frames(.data)) nframes <- nframes + 1
  paths <- align_paths(from, to, enter = enter, exit = exit, match = match)
  paths <- tween_state(paths$from, paths$to, ease = ease, nframes = nframes)
  paths <- paths[!paths$.frame %in% c(1, nframes), , drop = FALSE]
  if (is.character(paths$.id) || is.character(from$.id) || is.character(to$.id)) {
    paths$.id <- as.character(paths$.id)
    from$.id <- as.character(from$.id)
    to$.id <- as.character(to$.id)
  }
  morph <- vec_rbind(
    if (nframes > 1) cbind(from, .frame = rep(1, nrow(from))) else NULL,
    paths,
    cbind(to, .frame = rep(nframes, nrow(to)))
  )
  .with_prior_frames(.data, morph, nframes)
}

align_paths <- function(from, to, min_n = 50, enter, exit, match) {
  from <- make_paths(from)
  to <- make_paths(to)
  paths <- if (match) {
    prep_match_paths(from, to)
  } else {
    prep_align_paths(from, to)
  }
  paths <- mapply(
    match_shapes,
    from = paths$from,
    to = paths$to,
    MoreArgs = list(
      enter = enter, exit = exit, min_n = min_n, closed = FALSE
    ),
    SIMPLIFY = FALSE
  )
  from <- lapply(paths, `[[`, 'from')
  id <- rep(seq_along(from), vapply(from, nrow, integer(1)))
  from <- vec_rbind(!!!from)
  to <- vec_rbind(!!!lapply(paths, `[[`, 'to'))
  from$.id <- id
  to$.id <- id
  common_id(from = from, to = to)
}
prep_match_paths <- function(from, to) {
  all_ids <- as.character(union(names(from), names(to)))
  from_all <- structure(rep(list(NULL), length(all_ids)), names = all_ids)
  to_all <- from_all
  from_all[names(from)] <- from
  to_all[names(to)] <- to
  paths <- Map(function(from, to) {
    if (is.null(from) || is.null(to) || (length(from) == 1 && length(to) == 1)) {
      list(from = from, to = to)
    } else {
      divide_paths(from, to)
    }
  }, from = from_all, to = to_all)
  from_all <- lapply(paths, `[[`, 'from')
  to_all <- lapply(paths, `[[`, 'to')
  list(from = from_all, to = to_all)
}
prep_align_paths <- function(from, to) {
  from <- unlist(from, recursive = FALSE)
  to <- unlist(to, recursive = FALSE)
  paths <- divide_paths(from, to)
  from <- paths$from
  to <- paths$to
  names(from) <- as.character(seq_along(from))
  names(to) <- as.character(seq_along(to))
  list(from = lapply(from, list), to = lapply(to, list))
}
make_paths <- function(x) {
  lapply(split(x, x$.id), to_path)
}
to_path <- function(path) {
  gaps <- which(is.na(path$x) | is.na(path$y))
  if (length(gaps) == 0) {
    path <- list(path)
  } else {
    gaps <- c(gaps, nrow(path) + 1)
    path <- path[-gaps, , drop = FALSE]
    path <- split(path, rep(seq_along(gaps), diff(c(0, gaps)) - 1))
  }
  lapply(path, function(line) {
    if (nrow(line) == 1) {
      line[c(1,1), ]
    } else {
      line
    }
  })
}
#' @importFrom sf st_multilinestring st_linestring
to_line <- function(path) {
  if (is.data.frame(path)) {
    if (nrow(path) == 1) {
      st_linestring(cbind(path$x[c(1, 1)], path$y[c(1, 1)]))
    } else {
      st_linestring(cbind(path$x, path$y))
    }
  } else {
    st_multilinestring(lapply(path, function(x) {
      if (nrow(x) == 1) {
        cbind(x$x[c(1, 1)], x$y[c(1, 1)])
      } else {
        cbind(x$x, x$y)
      }
    }))
  }
}
cut_lines <- function(lines, lengths, n) {
  longest <- order(lengths, decreasing = TRUE)
  n_splits <- find_splits(lengths[longest], n)
  splits <- n_splits[match(seq_along(lengths), longest)] + 1
  lines <- Map(function(path, n) {
    if (n == 1) return(list(path))
    if (n > nrow(path) - 1) path <- add_points(path, n - nrow(path) + 1, FALSE)
    paths <- split(seq_len(nrow(path)), sort(seq_len(nrow(path))%%n))
    lapply(paths, function(i) {
      if (i[1] != 1) i <- c(i[1] - 1, i)
      path[i, , drop = FALSE]
    })
  }, path = lines, n = splits)
  unlist(lines, recursive = FALSE)
}
#' @importFrom sf st_length
#' @importFrom lpSolve lp.assign
divide_paths <- function(from, to) {
  from_st <- st_sfc(lapply(from, to_line))
  to_st <- st_sfc(lapply(to, to_line))
  if (length(from) < length(to)) {
    l <- st_length(from_st)
    from <- cut_lines(from, l, length(to) - length(from))
    from_st <- st_sfc(lapply(from, to_line))
  } else if (length(to) < length(from)) {
    l <- st_length(to_st)
    to <- cut_lines(to, l, length(from) - length(to))
    to_st <- st_sfc(lapply(to, to_line))
  }
  from_st <- st_normalize(from_st)
  to_st <- st_normalize(to_st)
  if (length(from_st) != 1 && length(to_st) != 1) {
    distance <- st_distance(st_centroid(from_st), st_centroid(to_st))
    length_diff <- abs(outer(st_length(from_st), st_length(to_st), `-`))
    if (max(length_diff) > 0) distance <- distance * (1 + length_diff / max(length_diff))
    match_poly <- lp.assign(distance)
    if (match_poly$status == 0) {
      to <- to[apply(round(match_poly$solution) == 1, 1, which)]
    }
  }
  list(from = from, to = to)
}

#' Some different geometries to play with
#'
#' These functions are provided to allow you to play with somee simple shapes as
#' you explore `transformr` and are also used in the examples for the different
#' tween functions. All geometries can be returned as either a standard
#' `data.frame` with `x`, `y`, and `id` column, or as an sf geometry of the
#' appropriate type.
#'
#' @param st Logical. Should the geometry be returned as an `sf` feature?
#' @param detail The number of points defining the shape
#' @param n For `poly_circles` the number of circles, for `poly_star` and
#' `poly_star_hole` the number of 'arms', and for `point_random` the number of
#' points
#' @param r,r1 The radius of the geometry. `r` gives the radius of the circles
#' in `poly_circles` and `r1` gives the inner radius for
#' `poly_star`/`poly_star_hole`, thus determining how pointy it is
#' @param windings The number of revolutions in the spiral
#' @param w1,w2 The frequency for the two sine waves
#' @param dim the number of rows and columns in the grid
#'
#' @return Either a data.frame or an sf feature depending on the value of `st`
#'
#' @name simple_shapes
#' @rdname simple_shapes
#'
#' @examples
#' # Create a 7-pointed star
#' poly_star(n = 7)
#'
NULL

#' @rdname simple_shapes
#' @export
poly_circle <- function(st = FALSE, detail = 360) {
  i <- seq(0, 2*pi, length.out = detail+1)[-detail-1]
  x <- sin(i)
  y = cos(i)
  if (st) {
    st_polygon(list(cbind(x, y)[c(seq_len(detail), 1), , drop = FALSE]))
  } else {
    data_frame(x = x, y = y, id = 1L)
  }
}
#' @rdname simple_shapes
#' @importFrom sf st_multipolygon
#' @export
poly_circles <- function(st = FALSE, n = 3, r = 0.25, detail = 360) {
  i <- seq(0, 2*pi, length.out = n+1)[-n-1]
  x <- sin(i)
  y = cos(i)
  d <- poly_circle(detail = detail)
  d_small <- d
  d_small$x <- d_small$x * r
  d_small$y <- d_small$y * r
  d1 <- d_small
  d2 <- d_small
  d3 <- d_small
  d1$x <- d1$x + x[1]
  d1$y <- d1$y + y[1]
  d2$x <- d2$x + x[2]
  d2$y <- d2$y + y[2]
  d2$id <- 2L
  d3$x <- d3$x + x[3]
  d3$y <- d3$y + y[3]
  d3$id <- 3L
  if (st) {
    st_multipolygon(list(
      list(cbind(d1$x, d1$y)[c(seq_len(360), 1), , drop = FALSE]),
      list(cbind(d2$x, d2$y)[c(seq_len(360), 1), , drop = FALSE]),
      list(cbind(d3$x, d3$y)[c(seq_len(360), 1), , drop = FALSE])
    ))
  } else {
    vec_rbind(d1,d2,d3)
  }
}
#' @rdname simple_shapes
#' @export
poly_star <- function(st = FALSE, n = 5, r1 = 0.5) {
  d <- poly_circle(detail = n*2)
  d$x[c(FALSE, TRUE)] <- d$x[c(FALSE, TRUE)] * r1
  d$y[c(FALSE, TRUE)] <- d$y[c(FALSE, TRUE)] * r1
  if (st) {
    d <- st_polygon(list(cbind(d$x, d$y)[c(seq_len(n*2), 1), , drop = FALSE]))
  }
  d
}
#' @rdname simple_shapes
#' @export
poly_star_hole <- function(st = FALSE, n = 5, r1 = 0.5) {
  d <- poly_star(n = n, r1 = r1)
  d1 <- d
  d1$x <- d1$x * 0.5
  d1$y <- d1$y * 0.5
  if (st) {
    st_polygon(list(
      cbind(d$x, d$y)[c(seq_len(n*2), 1), , drop = FALSE],
      cbind(d1$x, d1$y)[c(seq_len(n*2), 1), , drop = FALSE]
    ))
  } else {
    vec_rbind(
      d,
      data_frame(x = NA, y = NA, id = 1L),
      d1
    )
  }
}
#' @rdname simple_shapes
#' @export
path_spiral <- function(st = FALSE, windings = 5) {
  n = 50 * windings
  r = seq(0, 1, length.out = n)
  i <- seq(0, 2*pi*windings, length.out = n+1)[-n-1]
  x <- sin(i) * r
  y <- cos(i) * r
  if (st) {
    st_linestring(cbind(x, y))
  } else {
    data_frame(x = x, y = y, id = 1L)
  }
}
#' @rdname simple_shapes
#' @export
path_waves <- function(st = FALSE, w1 = 7, w2 = 11) {
  x <- seq(-1, 1, length.out = 150)
  y1 = 0.2*sin(w1*x) + 0.5
  y2 = 0.2*sin(w2*x) - 0.5
  if (st) {
    st_multilinestring(list(
      cbind(x, y1),
      cbind(x, y2)
    ))
  } else {
    data_frame(x = rep(x, 2), y = c(y1, y2), id = rep(c(1L, 2L), each = length(x)))
  }
}
#' @rdname simple_shapes
#' @export
#' @importFrom sf st_multipoint
#' @importFrom stats runif
point_random <- function(st = FALSE, n = 10) {
  x <- runif(10, min = -1, max = 1)
  y <- runif(10, min = -1, max = 1)
  if (st) {
    st_multipoint(cbind(x, y))
  } else {
    data_frame(x = x, y = y, i = seq_len(n))
  }
}
#' @rdname simple_shapes
#' @export
#' @importFrom sf st_multipoint
point_grid <- function(st = FALSE, dim = 5) {
  x <- rep(seq(-1, 1, length.out = dim), each = dim)
  y <- rep(seq(-1, 1, length.out = dim), dim)
  if (st) {
    st_multipoint(cbind(x, y))
  } else {
    data_frame(x = x, y = y, i = seq_len(dim^2))
  }
}

find_splits <- function(lengths, n) {
  find_splits_c(as.numeric(lengths), as.integer(n))
}
insert_points <- function(x, y, splits, n) {
  insert_points_c(as.numeric(x), as.numeric(y), as.integer(splits), as.integer(n))
}
rotate <- function(poly_x, poly_y, ref_x, ref_y) {
  rotate_c(as.numeric(poly_x), as.numeric(poly_y), as.numeric(ref_x), as.numeric(ref_y))
}

add_points <- function(polygon, n, connect = TRUE) {
  if (n < 1) {
    return(polygon)
  }
  if (nrow(polygon) == 1) {
    return(polygon[rep(1, n + 1), , drop = FALSE])
  }
  if (connect) {
    x <- diff(c(polygon$x, polygon$x[1]))
    y <- diff(c(polygon$y, polygon$y[1]))
  } else {
    x <- diff(polygon$x)
    y <- diff(polygon$y)
  }
  l <- sqrt(x*x + y*y)
  longest <- order(l, decreasing = TRUE)
  n_splits <- find_splits(l[longest], n)
  splits <- n_splits[match(seq_along(l), longest)]
  if (!connect) splits <- c(splits, 0)
  new_points <- insert_points(polygon$x, polygon$y, splits, n)
  new_polygon <- polygon[rep(seq_len(nrow(polygon)), splits + 1), ]
  new_polygon$x <- new_points$x
  new_polygon$y <- new_points$y
  if ('ymin' %in% names(polygon) && length(unique(polygon$ymin)) > 1) {
    new_points <- insert_points(polygon$x, polygon$ymin, splits, n)
    new_polygon$ymin <- new_points$y
  }
  if ('ymax' %in% names(polygon) && length(unique(polygon$ymax)) > 1) {
    new_points <- insert_points(polygon$x, polygon$ymax, splits, n)
    new_polygon$ymax <- new_points$y
  }
  if ('xmin' %in% names(polygon) && length(unique(polygon$xmin)) > 1) {
    new_points <- insert_points(polygon$xmin, polygon$y, splits, n)
    new_polygon$xmin <- new_points$x
  }
  if ('xmax' %in% names(polygon) && length(unique(polygon$xmax)) > 1) {
    new_points <- insert_points(polygon$xmax, polygon$y, splits, n)
    new_polygon$xmax <- new_points$x
  }
  new_polygon
}

#' @importFrom rlang as_function
match_shapes <- function(from, to, enter, exit, min_n, closed = TRUE) {
  if (is.null(from)) {
    if (is.null(enter)) {
      from <- list(to[[1]][0, , drop = FALSE])
      to <- from
    } else {
      from <- lapply(to, function(x) {
        x <- as_function(enter)(x)
        x$.phase <- 'enter'
        x
      })
    }
  } else if (is.null(to)) {
    if (is.null(exit)) {
      to <- list(from[[1]][0, , drop = FALSE])
      from <- to
    } else {
      to <- lapply(from, function(x) {
        x <- as_function(exit)(x)
        x$.phase <- 'exit'
        x
      })
    }
  } else {
    matched <- if (closed) match_polygon(from, to, min_n) else match_path(from, to, min_n)
    from <- matched$from
    to <- matched$to
  }
  from <- vec_rbind(!!!lapply(from, function(x) x[seq_len(nrow(x) + 1), ]))
  to <- vec_rbind(!!!lapply(to, function(x) x[seq_len(nrow(x) + 1), ]))
  from <- from[-nrow(from), ]
  to <- to[-nrow(to), ]
  list(from = from, to = to)
}
match_polygon <- function(from, to, min_n) {
  main_from <- from[[1]]
  main_to <- to[[1]]
  n_points <- max(c(min_n, nrow(main_from), nrow(main_to)))
  if (nrow(main_from) < n_points) main_from <- add_points(main_from, n_points - nrow(main_from), connect = TRUE)
  if (nrow(main_to) < n_points) main_to <- add_points(main_to, n_points - nrow(main_to), connect = TRUE)
  offset <- rotate(main_to$x, main_to$y, main_from$x, main_from$y)
  to_end <- seq_len(nrow(main_to)) < offset
  main_to <- vec_rbind(main_to[!to_end, , drop = FALSE],
                       main_to[to_end, , drop = FALSE])
  from[[1]] <- main_from
  to[[1]] <- main_to
  if (length(from) > 1 || length(to) > 1) {
    holes <- align_holes(from, to)
    from <- c(from[1], holes$from)
    to <- c(to[1], holes$to)
  }
  list(from = from, to = to)
}
match_path <- function(from, to, min_n) {
  paths <- Map(function(from, to) {
    n_points <- max(c(min_n, nrow(from), nrow(to)))
    if (nrow(from) < n_points) from <- add_points(from, n_points - nrow(from), connect = FALSE)
    if (nrow(to) < n_points) to <- add_points(to, n_points - nrow(to), connect = FALSE)
    list(from = from, to = to)
  }, from = from, to = to)
  list(from = lapply(paths, `[[`, 'from'), to = lapply(paths, `[[`, 'to'))
}
common_id <- function(from, to) {
  from_id <- as.character(from$.id)
  to_id <- as.character(to$.id)
  all_id <- unique(c(from_id, to_id))
  from$.id <- match(from_id, all_id)
  to$.id <- match(to_id, all_id)
  list(from = from, to = to)
}

first <- function(x) x[[1]]
`first<-` <- function(x, value) {
  x[[1]] <- value
  x
}
last <- function(x) x[[length(x)]]
`last<-` <- function(x, value) {
  x[[length(x)]] <- value
  x
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

find_splits <- function(lengths, n) {
    .Call('_transformr_find_splits', PACKAGE = 'transformr', lengths, n)
}

insert_points <- function(x, y, splits, n) {
    .Call('_transformr_insert_points', PACKAGE = 'transformr', x, y, splits, n)
}

rotate <- function(poly_x, poly_y, ref_x, ref_y) {
    .Call('_transformr_rotate', PACKAGE = 'transformr', poly_x, poly_y, ref_x, ref_y)
}

fill_down <- function(nas) {
    .Call('_transformr_fill_down', PACKAGE = 'transformr', nas)
}

unpack_sf <- function(sf, type) {
    .Call('_transformr_unpack_sf', PACKAGE = 'transformr', sf, type)
}

repack_sf <- function(df, type, n_frames) {
    .Call('_transformr_repack_sf', PACKAGE = 'transformr', df, type, n_frames)
}

# Generated by cpp11: do not edit by hand

find_splits_c <- function(lengths, n) {
  .Call(`_transformr_find_splits_c`, lengths, n)
}

insert_points_c <- function(x, y, splits, n) {
  .Call(`_transformr_insert_points_c`, x, y, splits, n)
}

rotate_c <- function(poly_x, poly_y, ref_x, ref_y) {
  .Call(`_transformr_rotate_c`, poly_x, poly_y, ref_x, ref_y)
}

fill_down <- function(nas) {
  .Call(`_transformr_fill_down`, nas)
}

unpack_sf <- function(sf, type) {
  .Call(`_transformr_unpack_sf`, sf, type)
}

repack_sf <- function(df, type, n_frames) {
  .Call(`_transformr_repack_sf`, df, type, n_frames)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shapes.R
\name{simple_shapes}
\alias{simple_shapes}
\alias{poly_circle}
\alias{poly_circles}
\alias{poly_star}
\alias{poly_star_hole}
\alias{path_spiral}
\alias{path_waves}
\alias{point_random}
\alias{point_grid}
\title{Some different geometries to play with}
\usage{
poly_circle(st = FALSE, detail = 360)

poly_circles(st = FALSE, n = 3, r = 0.25, detail = 360)

poly_star(st = FALSE, n = 5, r1 = 0.5)

poly_star_hole(st = FALSE, n = 5, r1 = 0.5)

path_spiral(st = FALSE, windings = 5)

path_waves(st = FALSE, w1 = 7, w2 = 11)

point_random(st = FALSE, n = 10)

point_grid(st = FALSE, dim = 5)
}
\arguments{
\item{st}{Logical. Should the geometry be returned as an \code{sf} feature?}

\item{detail}{The number of points defining the shape}

\item{n}{For \code{poly_circles} the number of circles, for \code{poly_star} and
\code{poly_star_hole} the number of 'arms', and for \code{point_random} the number of
points}

\item{r, r1}{The radius of the geometry. \code{r} gives the radius of the circles
in \code{poly_circles} and \code{r1} gives the inner radius for
\code{poly_star}/\code{poly_star_hole}, thus determining how pointy it is}

\item{windings}{The number of revolutions in the spiral}

\item{w1, w2}{The frequency for the two sine waves}

\item{dim}{the number of rows and columns in the grid}
}
\value{
Either a data.frame or an sf feature depending on the value of \code{st}
}
\description{
These functions are provided to allow you to play with somee simple shapes as
you explore \code{transformr} and are also used in the examples for the different
tween functions. All geometries can be returned as either a standard
\code{data.frame} with \code{x}, \code{y}, and \code{id} column, or as an sf geometry of the
appropriate type.
}
\examples{
# Create a 7-pointed star
poly_star(n = 7)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transformr-package.R
\docType{package}
\name{transformr-package}
\alias{transformr}
\alias{transformr-package}
\title{transformr: Polygon and Path Transformations}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

In order to smoothly animate the transformation of polygons and paths, many aspects needs to be taken into account, such as differing number of control points, changing center of rotation, etc. The 'transformr' package provides an extensive framework for manipulating the shapes of polygons and paths and can be seen as the spatial brother to the 'tweenr' package.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/thomasp85/transformr}
  \item Report bugs at \url{https://github.com/thomasp85/transformr/issues}
}

}
\author{
\strong{Maintainer}: Thomas Lin Pedersen \email{thomasp85@gmail.com} (\href{https://orcid.org/0000-0002-5147-4711}{ORCID})

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tween_sf.R
\name{tween_sf}
\alias{tween_sf}
\title{Transition between data.frames containing sfc columns}
\usage{
tween_sf(.data, to, ease, nframes, id = NULL, enter = NULL, exit = NULL)
}
\arguments{
\item{.data}{A data.frame to start from. If \code{.data} is the result of a prior
tween, only the last frame will be used for the tween. The new tween will
then be added to the prior tween}

\item{to}{A data.frame to end at. It must contain the same columns as .data
(exluding \code{.frame})}

\item{ease}{The easing function to use. Either a single string or one for
each column in the data set.}

\item{nframes}{The number of frames to calculate for the tween}

\item{id}{The column to match observations on. If \code{NULL} observations will be
matched by position. See the \emph{Match, Enter, and Exit} section for more
information.}

\item{enter}{functions that calculate a start state for new observations
that appear in \code{to} or an end state for observations that are not present in
\code{to}. If \code{NULL} the new/old observations will not be part of the tween. The
function gets a data.frame with either the start state of the exiting
observations, or the end state of the entering observations and must return
a modified version of that data.frame. See the \emph{Match, Enter, and Exit}
section for more information.}

\item{exit}{functions that calculate a start state for new observations
that appear in \code{to} or an end state for observations that are not present in
\code{to}. If \code{NULL} the new/old observations will not be part of the tween. The
function gets a data.frame with either the start state of the exiting
observations, or the end state of the entering observations and must return
a modified version of that data.frame. See the \emph{Match, Enter, and Exit}
section for more information.}
}
\value{
A data.frame containing intermediary states
}
\description{
This function is equivalent to \code{\link[tweenr:tween_state]{tweenr::tween_state()}} except that it
understands \code{\link[sf:sfc]{sf::sfc}} columns, as defined by the \code{sf} package. An \code{sfc}
column is a column containing simple features and can this hold both points,
lines polygons and more. \code{tween_sf} currently has support for (multi)point,
(multi)linestring, and (multi)polygon types and requires that the transition
is between compatible types (points-to-points, linestring-to-linestring,
polygon-to-polygon). For (multi)linestring and (multi)polygon, the behavior
is similar to \code{\link[=tween_path]{tween_path()}} and \code{\link[=tween_polygon]{tween_polygon()}} respectively, with each
feature beeing run through the respective function with \code{match = FALSE}. For
(multi)points it behaves more or less like \code{\link[tweenr:tween_state]{tweenr::tween_state()}} except
additional points are added as needed to make the to stages contain the same
number of points. Points are added on top of existing points so it appears as
if the points are divided into more.
}
\examples{
library(magrittr)
star_hole <- poly_star_hole(st = TRUE)
circles <- poly_circles(st = TRUE)
spiral <- path_spiral(st = TRUE)
waves <- path_waves(st = TRUE)
random <- point_random(st = TRUE)
grid <- point_grid(st = TRUE)
df1 <- data.frame(geo = sf::st_sfc(star_hole, spiral, random))
df2 <- data.frame(geo = sf::st_sfc(circles, waves, grid))

tween_sf(df1, df2, 'linear', 30)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tween_path.R
\name{tween_path}
\alias{tween_path}
\title{Transition between path data frames}
\usage{
tween_path(
  .data,
  to,
  ease,
  nframes,
  id = NULL,
  enter = NULL,
  exit = NULL,
  match = TRUE
)
}
\arguments{
\item{.data}{A data.frame to start from. If \code{.data} is the result of a prior
tween, only the last frame will be used for the tween. The new tween will
then be added to the prior tween}

\item{to}{A data.frame to end at. It must contain the same columns as .data
(exluding \code{.frame})}

\item{ease}{The easing function to use. Either a single string or one for
each column in the data set.}

\item{nframes}{The number of frames to calculate for the tween}

\item{id}{The column to match observations on. If \code{NULL} observations will be
matched by position. See the \emph{Match, Enter, and Exit} section for more
information.}

\item{enter, exit}{functions that calculate a start state for new observations
that appear in \code{to} or an end state for observations that are not present in
\code{to}. If \code{NULL} the new/old observations will not be part of the tween. The
function gets a data.frame with either the start state of the exiting
observations, or the end state of the entering observations and must return
a modified version of that data.frame. See the \emph{Match, Enter, and Exit}
section for more information.}

\item{match}{Should polygons be matched by id? If \code{FALSE} then polygons will
be matched by shortest distance and if any state has more polygons than the
other, the other states polygons will be chopped up so the numbers match.}
}
\value{
A data.frame containing intermediary states
}
\description{
This function is equivalent to \code{\link[tweenr:tween_state]{tweenr::tween_state()}} but expects the data
to have an \code{x} and \code{y} column and encode paths.
}
\section{Aligning paths}{

There is less work required to align paths than there is to align polygons,
simply because no rotation is possible/required, and the notion of clockwise
winding order is not meaningful in the scope of paths. Still, paths need to
be matched and the number of points in each pair of matched paths must be
equal. Paths are matched based on relative length rather than on position and
seek to minimize the change in length during transition. This is chosen from
the point of view that huge elongation or contraction are much more
distracting than longer travel distances.
}

\section{Cutting paths}{

If the number of paths to transition between is not even, some of the paths
need to be cut in order to succesfully match the paths. The cuts are
distributed based on the same algorithm that distributes cuts in polygons and
seek to cut the lines into as even-length pieces as possible.
}

\section{Multipaths}{

It is possible to encode multiple paths with the same id be separating them
with a \code{NA} row, much in the same way as holes are encoded in polygons. If
paths are not matched based on id (\code{match = FALSE}) then multipaths will
simply be split into separate paths. On the other hand, if paths are matched
by id all paths within a multipath will transition into the (multi)path that
has the same id in the other state.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tween_polygon.R
\name{st_normalize}
\alias{st_normalize}
\title{Normalise a geometry to fit inside a unit square}
\usage{
st_normalize(st)
}
\arguments{
\item{st}{An sf geometry such as \code{sf}, \code{sfc}, or \code{sfg}}
}
\value{
An object of the same type as \code{st}
}
\description{
This is a small helper function that will take an sf geometry and fit it
inside the unit square (a square centered on 0 and ranging from -1 to 1 in
both dimensions). The function will retain the aspect ratio of the geometry
and simply scale it down until it fits.
}
\examples{
library(sf)
nc <- st_read(system.file("shape/nc.shp", package="sf"), quiet = TRUE)
st_bbox(nc)

nc_norm <- st_normalize(nc)
st_bbox(nc_norm)

}
